"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var sourceMap = require("source-map");
var messages_formatter_1 = require("../../lib/messages-formatter");
var GLOBAL_ISTANBUL_CONTAINER_NAMES = ["__coverage__"];
var IstanbulToFootprints = /** @class */ (function () {
    function IstanbulToFootprints(cfg, logger) {
        this.cfg = cfg;
        this.totalTests = 0;
        this.previousIstanbulCounters = {};
        this.relativeModulePaths = {};
        if (!logger)
            throw new Error("logger is required");
        if (!cfg)
            throw new Error("cfg is required");
        this.logger = logger;
    }
    IstanbulToFootprints.prototype.resetState = function () {
        this.totalTests = 0;
        //Holds the a mapping between a uniqueId for an element in a test, to an index in this elements 'hits' array.
        this.eidToHitsIndex = {};
        this.uniqueIdToElement = {};
        this.testNameToTestData = {};
        this.fileNameToAppFile = {};
        this.fileToSourceMapConsumer = {};
    };
    IstanbulToFootprints.prototype.createFootprintsPackageFromQueue = function (queuedFootprints, extraData) {
        var _this = this;
        this.resetState();
        var result = this.createFootprintsAppData(extraData);
        var istanbulMaps = this.getIstanbulCounters();
        queuedFootprints.forEach(function (fp) {
            _this.convertIstanbulDataToFootprints(fp, result, istanbulMaps);
        });
        return result;
    };
    IstanbulToFootprints.prototype.createFootprintsPackageFromRawData = function (rawData, extraData) {
        this.resetState();
        var result = this.createFootprintsAppData(extraData);
        this.convertIstanbulDataToFootprints(rawData, result);
        return result;
    };
    IstanbulToFootprints.prototype.convertIstanbulDataToFootprints = function (footprintsElement, result, istanbulMaps) {
        var test = this.getOrCreateTestData(footprintsElement.testName, footprintsElement.executionId, footprintsElement.localTime, result);
        var app = result.apps[0]; //We assume single App for now.
        var fullIstanbulData = istanbulMaps || footprintsElement.footprints;
        for (var moduleName in footprintsElement.footprints) {
            var istanbulModule = footprintsElement.footprints[moduleName];
            var moduleWithMaps = fullIstanbulData[moduleName];
            this.logger.debug("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
            this.logger.debug("moduleName: ", moduleName);
            if (!moduleWithMaps) {
                this.logger.info("Cannot find full istanbul data for %s", moduleName);
                continue;
            }
            var relPath = this.getRelativeModulePath(moduleWithMaps);
            var sourceMap_1 = this.getSourceMapForFile(relPath);
            var file = this.getOrCreateFootprintsAppFile(relPath, app);
            if (istanbulModule.f) {
                this.handleFunctionsData(istanbulModule, moduleWithMaps, sourceMap_1, relPath, file, test);
            }
            if (istanbulModule.b) {
                this.handleBranchesData(istanbulModule, moduleWithMaps, sourceMap_1, relPath, file, test);
            }
        }
    };
    IstanbulToFootprints.prototype.handleBranchesData = function (istanbulModule, moduleWithMaps, sourceMap, relPath, file, test) {
        for (var branchId in istanbulModule.b) {
            var ifElseAndSwitchCasesArr = istanbulModule.b[branchId];
            var expressionInfo = moduleWithMaps.branchMap[branchId];
            for (var idx = 0; idx < ifElseAndSwitchCasesArr.length; idx++) {
                var branchInfo = expressionInfo.locations[idx];
                var hits = ifElseAndSwitchCasesArr[idx];
                if (hits > 0) {
                    var start = branchInfo.start;
                    var sourceMapData = this.readSourceMapData(sourceMap, start, relPath);
                    var filePath = relPath;
                    if (sourceMapData) {
                        filePath = sourceMapData.originalFilename;
                        start = sourceMapData.originalStart;
                    }
                    var uniqueId = filePath + "|" + this.formatLoc(start) + "|" + idx;
                    var branchElement = this.getOrCreateBranchElement(uniqueId, file);
                    if (this.cfg.extendedFootprints.value) {
                        branchElement.branchInfo = JSON.stringify(branchInfo);
                        branchElement.sourceMapData = JSON.stringify(sourceMapData);
                    }
                    this.addOrUpdateElementHits(test, uniqueId, branchElement, hits);
                }
            }
        }
    };
    IstanbulToFootprints.prototype.handleFunctionsData = function (istanbulModule, moduleWithMaps, sourceMap, relPath, file, test) {
        for (var funcId in istanbulModule.f) {
            var hits = istanbulModule.f[funcId];
            if (hits > 0) {
                var funcInfo = moduleWithMaps.fnMap[funcId];
                var start = funcInfo.loc.start;
                var sourceMapData = this.readSourceMapData(sourceMap, start, relPath);
                var filePath = relPath;
                if (sourceMapData) {
                    filePath = sourceMapData.originalFilename;
                    start = sourceMapData.originalStart;
                }
                var uniqueId = filePath + "@" + this.formatLoc(start);
                var methodElement = this.getOrCreateMethodElement(uniqueId, file);
                if (this.cfg.extendedFootprints.value) {
                    methodElement.funcInfo = JSON.stringify(funcInfo);
                    methodElement.sourceMapData = JSON.stringify(sourceMapData);
                }
                this.addOrUpdateElementHits(test, uniqueId, methodElement, hits);
                if (funcInfo.decl) {
                    this.addDeclarationFootprints(funcInfo, sourceMap, relPath, filePath, file, test, hits);
                }
            }
        }
    };
    IstanbulToFootprints.prototype.addDeclarationFootprints = function (funcInfo, sourceMap, relPath, filePath, file, test, hits) {
        var declStart = funcInfo.decl.start; //For NYC
        var sourceMapData = this.readSourceMapData(sourceMap, declStart, relPath);
        if (sourceMapData) {
            filePath = sourceMapData.originalFilename;
            declStart = sourceMapData.originalStart;
        }
        var declUniqueId = filePath + "@" + this.formatLoc(declStart);
        var declMethodElement = this.getOrCreateMethodElement(declUniqueId, file);
        this.addOrUpdateElementHits(test, declUniqueId, declMethodElement, hits);
    };
    IstanbulToFootprints.prototype.getSourceMapForFile = function (relPath) {
        if (!this.fileToSourceMapConsumer[relPath]) {
            this.fileToSourceMapConsumer[relPath] = this.readSourceMaps(relPath);
            this.logger.debug("Read source maps for file':" + relPath + "'");
        }
        var sourceMap = this.fileToSourceMapConsumer[relPath];
        return sourceMap;
    };
    IstanbulToFootprints.prototype.addOrUpdateElementHits = function (testData, elementUniqueId, element, hits) {
        var HITS_INDEX = 1;
        var eid = elementUniqueId + "|" + testData.index;
        var index = this.eidToHitsIndex[eid];
        if (index != null) {
            var testIndexAndHits = element.hits[index];
            var elementHits = testIndexAndHits[HITS_INDEX];
            elementHits = elementHits + hits;
            testIndexAndHits[HITS_INDEX] = elementHits;
        }
        else {
            var testIndexAndHits = new Array();
            testIndexAndHits.push(testData.index);
            testIndexAndHits.push(hits);
            var length = element.hits.push(testIndexAndHits);
            this.eidToHitsIndex[eid] = length - 1;
        }
    };
    IstanbulToFootprints.prototype.getOrCreateTestData = function (testName, executionId, localTime, result) {
        if (!this.testNameToTestData[testName]) {
            var test_1 = {
                executionId: executionId,
                localTime: localTime,
                testName: testName
            };
            result.tests.push(test_1);
            this.testNameToTestData[testName] = {
                index: this.totalTests++,
                test: test_1
            };
        }
        return this.testNameToTestData[testName];
    };
    IstanbulToFootprints.prototype.getOrCreateMethodElement = function (uniqueId, file) {
        if (!this.uniqueIdToElement[uniqueId]) {
            var x = {
                hits: new Array(),
                uniqueId: uniqueId
            };
            this.uniqueIdToElement[uniqueId] = x;
            file.methods.push(x);
        }
        return this.uniqueIdToElement[uniqueId];
    };
    IstanbulToFootprints.prototype.getOrCreateBranchElement = function (uniqueId, file) {
        if (!this.uniqueIdToElement[uniqueId]) {
            var x = {
                hits: new Array(),
                uniqueId: uniqueId
            };
            this.uniqueIdToElement[uniqueId] = x;
            file.branches.push(x);
        }
        return this.uniqueIdToElement[uniqueId];
    };
    IstanbulToFootprints.prototype.getOrCreateFootprintsAppFile = function (fileName, app) {
        if (this.fileNameToAppFile[fileName]) {
            return this.fileNameToAppFile[fileName];
        }
        //File doesn't exist in request.
        var file = {
            methods: new Array(),
            branches: new Array(),
            lines: new Array(),
            path: fileName
        };
        app.files.push(file);
        this.fileNameToAppFile[fileName] = file;
        return file;
    };
    IstanbulToFootprints.prototype.createFootprintsAppData = function (extraData) {
        //Supporting only one app right now
        var app = {
            appName: this.cfg.appName.value,
            branchName: this.cfg.branch.value,
            buildName: this.cfg.build.value,
            moduleName: null,
            files: new Array()
        };
        extraData.apps.push(app);
        return extraData;
    };
    IstanbulToFootprints.prototype.readSourceMaps = function (fullFilename) {
        if (!fullFilename)
            return;
        var sourceMapsFilename = fullFilename + '.map';
        var contents, rawSourceMapJsonData, consumer;
        try {
            contents = fs.readFileSync(sourceMapsFilename).toString();
        }
        catch (e) {
            //Sourcemap file not found
            return null;
        }
        try {
            rawSourceMapJsonData = JSON.parse(contents);
        }
        catch (e) {
            console.error('[Sealights] Ignoring invalid source map: ' + sourceMapsFilename);
            return null;
        }
        try {
            //console.log('Procssing source maps from file: '+sourceMapsFilename);
            consumer = new sourceMap.SourceMapConsumer(rawSourceMapJsonData);
            return consumer;
        }
        catch (e) {
            console.error('[Sealights] Error processing source maps for file: ' + sourceMapsFilename);
            return null;
        }
    };
    IstanbulToFootprints.prototype.readSourceMapData = function (sourceMaps, start, coveragePath) {
        if (!sourceMaps)
            return null;
        try {
            //console.log('Checking original position for '+name+' at'+JSON.stringify(start));
            var originalPosition = sourceMaps.originalPositionFor(start); //{ line: start.line, column: start[1] });
            //console.log('Original position: '+JSON.stringify(originalPosition));
            if (originalPosition && originalPosition.source && originalPosition.line !== null && originalPosition.column !== null) {
                var originalFilename = originalPosition.source;
                if (originalFilename.indexOf("webpack:///webpack/") === -1 &&
                    originalFilename.indexOf("webpack:///external") === -1) {
                    if (originalFilename.indexOf('webpack:///') === 0) {
                        originalFilename = originalFilename.substring(11); //'webpack:///'.length
                    }
                    originalFilename = this.fixRelPath(coveragePath, originalFilename);
                    var originalStart = { line: originalPosition.line, column: originalPosition.column };
                    var result = { originalFilename: originalFilename, originalStart: originalStart };
                    return result;
                }
            }
        }
        catch (e) {
            console.error(e);
        }
        return null;
    };
    IstanbulToFootprints.prototype.fixRelPath = function (p1, p2) {
        try {
            if (!p1 || p1.length == 0)
                return p1;
            if (!p2 || p2.length == 0)
                return p1;
            //Make sure both paths are unix-style
            p1 = this.adjustPathSlashes(p1);
            p2 = this.adjustPathSlashes(p2);
            if (p2[0] == '/')
                return p2; //Absolute path
            var parts1 = p1.split('/');
            parts1.pop(); //remove the filename
            var parts2 = p2.split('/');
            var p3 = parts1.concat(parts2);
            var finalPath = this.adjustPathSlashes(path.normalize(p3.join('/')));
            return finalPath;
        }
        catch (err) {
            //log.error("Failed to fix real path for source maps. Error: " + err + ". Original File Name: " + p1 + ". Function: " + p2);
            return p2;
        }
    };
    IstanbulToFootprints.prototype.adjustPathSlashes = function (filePath) {
        filePath = (filePath || "").replace(/\\/g, '/');
        return filePath;
    };
    IstanbulToFootprints.prototype.getIstanbulCounters = function () {
        if (this.globalCoverageContainer)
            return this.globalCoverageContainer;
        var re = /^\$\$cov_[0-9]+\$\$$/;
        var keys = Object.keys(global);
        var m;
        for (var i = 0; i < keys.length; i++) {
            var k = keys[i];
            if ((m = re.exec(k)) !== null || GLOBAL_ISTANBUL_CONTAINER_NAMES.indexOf(k) >= 0) {
                this.globalCoverageContainer = global[k];
                return this.globalCoverageContainer;
            }
        }
        return null;
    };
    IstanbulToFootprints.prototype.getRelativeModulePath = function (module) {
        if (!this.relativeModulePaths[module.path]) {
            this.relativeModulePaths[module.path] = path.relative(process.cwd(), module.path).replace(/\\/g, '/');
        }
        return this.relativeModulePaths[module.path];
    };
    IstanbulToFootprints.prototype.formatLoc = function (loc) {
        return loc.line + ',' + loc.column;
    };
    IstanbulToFootprints.prototype.getCurrentFootprints = function () {
        var diff = this.getIstanbulCountersDiff(this.getIstanbulCounters());
        return diff;
    };
    IstanbulToFootprints.prototype.getIstanbulCountersDiff = function (currentCounters) {
        if (!currentCounters) {
            this.logger.error("Couldn't get istanbul counters, return empty diff.");
            return {};
        }
        var diff = {};
        var diffModule = null;
        function createModuleForDiff(istanbulModule) {
            diffModule = diff[istanbulModule] = { b: {}, f: {}, s: {} };
            return diffModule;
        }
        var prevCounters = this.previousIstanbulCounters;
        for (var istanbulModule in currentCounters) {
            diffModule = null;
            if (!prevCounters[istanbulModule]) {
                prevCounters[istanbulModule] = { b: {}, f: {}, s: {} };
            }
            var prevModule = prevCounters[istanbulModule];
            for (var funcId in currentCounters[istanbulModule].f) {
                if (prevModule.f[funcId] === undefined) {
                    prevModule.f[funcId] = 0;
                }
                var currentFunctionHits = currentCounters[istanbulModule].f[funcId];
                var prevFunctionHits = prevModule.f[funcId] || 0;
                if (currentFunctionHits > prevFunctionHits) {
                    //this is a diff
                    diffModule = diffModule || createModuleForDiff(istanbulModule);
                    diffModule.f[funcId] = (currentFunctionHits - prevFunctionHits);
                    prevModule.f[funcId] = currentFunctionHits;
                }
            }
            for (var branchId in currentCounters[istanbulModule].b) {
                var currentBranchesHits = currentCounters[istanbulModule].b[branchId];
                var prevBranchHits = prevModule.b[branchId];
                if (prevBranchHits) {
                    for (var i = 0, j = currentBranchesHits.length; i < j; i++) {
                        var currentHits = currentBranchesHits[i], previousHits = prevBranchHits[i] || 0;
                        if (currentHits > previousHits) {
                            //this is a diff
                            diffModule = diffModule || createModuleForDiff(istanbulModule);
                            diffModule.b[branchId] = diffModule.b[branchId] || this.newNumberArray(currentBranchesHits.length);
                            diffModule.b[branchId][i] = currentHits - previousHits;
                            prevBranchHits[i] = currentHits;
                        }
                    }
                }
                else {
                    prevModule.b[branchId] = currentBranchesHits.slice();
                    var branchSetHasHits = false;
                    for (var i = 0, j = currentBranchesHits.length; i < j && !branchSetHasHits; i++) {
                        if (currentBranchesHits[i])
                            branchSetHasHits = true;
                    }
                    if (branchSetHasHits) {
                        diffModule = diffModule || createModuleForDiff(istanbulModule);
                        diffModule.b[branchId] = currentBranchesHits.slice();
                    }
                }
            }
        }
        this.logDiff(diff);
        return diff;
    };
    IstanbulToFootprints.prototype.logDiff = function (diff) {
        if (!this.logger.debug()) {
            return;
        }
        var modules = Object.keys(diff);
        var formatter = new messages_formatter_1.MessagesFormatter();
        modules.forEach(function (module) {
            var methods = diff[module].f || {};
            var branches = diff[module].b || {};
            var methodsSize = Object.keys(methods).length;
            var branchesSize = Object.keys(branches).length;
            formatter.addContent(module + ": " + methodsSize + " methods and " + branchesSize + " branches");
        });
        this.logger.debug("Found coverage data for %s files:\n %s", modules.length, formatter.content);
    };
    IstanbulToFootprints.prototype.newNumberArray = function (length) {
        var arr = [];
        for (var i = 0; i < length; i++)
            arr.push(0);
        return arr;
    };
    return IstanbulToFootprints;
}());
exports.IstanbulToFootprints = IstanbulToFootprints;
//# sourceMappingURL=istanbul-to-footprints.js.map